GridController = class('GridController', EntityController)
GridController.instance = nil
-- requires
require "GridPathfinder"
require "GridRender"
--

function getGridController()
    if not GridController.instance then GridController:new() end
    return GridController.instance
end

function GridController:initialize()
    EntityController.initialize( self )
    
    if GridController.instance then error("Cannot create multple instances of a singleton") end
    GridController.instance = self
    
    self.hexagonShape = Hexagon:new( 600 )
    self.systemSpacing = 2
    self.systemUnderMouse = false
    self.systems = {}
    self:generateGrid( 18, 10 )
    self.gridRender = GridRender( self )
    self.pathfinder = GridPathfinder()
end

function GridController:destroy()
    EntityController.destroy( self )
    GridController.instance = nil
    self.gridRender:destroy()
    self = nil
end

function GridController:update( dt )
    self.pathfinder:update( dt )
    local worldMousePos = getCurrentCamera():mousepos()
    self.systemUnderMouse = nil
    for i, system in ipairs(self.systems) do
        if system:isPointInside( worldMousePos.x, worldMousePos.y ) then
            self.systemUnderMouse = system
        end
    end
end

function GridController:systemAtCoords( i, j )
    if i > self.width or j > self.height or i < 0 or j < 0 then return false end
    local cell = self.grid[i][j]
    if cell and cell.enabled and cell.system then
        return cell.system
    else
        return false
    end
end

--[[ Cell Rules.
    Each cell must have at least one connected side.
    A connected side must have at least one port, on average two.
]]
function GridController:generateGrid( width, height )
    -- initalize the multidimensional grid array.
    self.grid = {}
    self.width = width
    self.height = height
    for i = 0, width do
        self.grid[i] = {}
        for j = 0, height do
            self.grid[i][j] = {
                enabled = true,
                system = nil
            }
        end
    end
    -- randomly disable some cells.
    for i = 0, width do
        for j = 0, height do
            local cell = self.grid[i][j]
            local dice = math.random( 5 )
            if dice == 1 then
                cell.enabled = false
            end
        end
    end
    -- populate enabled cells with a system.
    for i = 0, width do
        for j = 0, height do
            local cell = self.grid[i][j]
            if cell.enabled then
                cell.system = SystemController:new( self, i, j, self.hexagonShape, self.systemSpacing )
                self.systems[#self.systems+1] = cell.system
            end
        end
    end
    -- create ports.
    for i = 0, width do
        for j = 0, height do
            local cell = self.grid[i][j]
            if cell.enabled and cell.system then
                -- get neighbors.
                local neighbors = cell.system:listOfNeighbors()
                table.shuffle( neighbors )
                if #neighbors > 0 then
                    for n = 1, math.random(1, #neighbors) do
                        local neighbor = neighbors[n]
                        if not cell.system.usedSides[neighbor.side] and math.random(n) == 1 then
                            for n1 = 1, math.random(1,4) do
                                if math.random(n1) == 1 then
                                    cell.system:createPort( neighbor.system, neighbor.side )
                                end
                            end
                        end
                    end
                end
            end
        end
    end
end


function GridController:findNeighborSystem( i, j, side )
    local isEven = (i % 2 == 0)

    if isEven then
        if side == DIR_N then
            j = j - 1
        elseif side == DIR_S then
            j = j + 1
        elseif side == DIR_NW then
            i = i + 1
        elseif side == DIR_NE then
            i = i - 1
        elseif side == DIR_SW then
            i = i + 1
            j = j + 1
        elseif side == DIR_SE then
            i = i - 1
            j = j + 1
        end
    else
        if side == DIR_N then
            j = j - 1
        elseif side == DIR_S then
            j = j + 1
        elseif side == DIR_NW then
            i = i + 1
            j = j - 1
        elseif side == DIR_NE then
            i = i - 1
            j = j - 1
        elseif side == DIR_SW then
            i = i + 1
        elseif side == DIR_SE then
            i = i - 1
        end
    end

    if i < 0 or j < 0 or i > self.width or j > self.height then
        return false
    end
    
    local cell = self.grid[i][j]
    if cell.enabled and cell.system then
        return cell.system
    else
        return self:findNeighborSystem( i, j, side )
    end
    
    return false
end

function GridController:getSystemUnderMouse()
    return self.systemUnderMouse
end

function GridController:setSystemUnderMouse( systemController )
    self.systemUnderMouse = systemController
end

function GridController:isSystemUnderMouse( systemController )
    return self.systemUnderMouse == systemController
end