GridPathfinder = class('GridPathfinder', nil)
local _findPath
local MAX_CYCLES = 1000
local MAX_CYCLES_PER_RESUME = 8

function GridPathfinder:initialize()
    self.routines = {}
end

function GridPathfinder:findPath( startSystem, endSystem, callback )
    local routine = coroutine.create( _findPath )
    table.insert( self.routines, routine )
    coroutine.resume( routine, startSystem, endSystem, callback )
end

function GridPathfinder:update( dt )
    -- resume suspended coroutines.
    for i, routine in ipairs(self.routines) do
        if coroutine.status( routine ) == "suspended" then
            coroutine.resume( routine )
        end
    end
    -- remove dead coroutines, we iterate backwards over the table
    -- because we might be removing multiple elements.
    for i = #self.routines, 1, -1 do
        if coroutine.status( self.routines[i] ) == "dead" then
            table.remove( self.routines, i )
        end
    end
end

--[[
create a list P
add the start node S, to P giving it one element
Until first path of P ends with G, or P is empty
       extract the first path from P
       extend first path one step to all neighbors creating X new paths
       reject all paths with loops
               for all paths that end at the same node, keep only the shortest one.
       add each remaining new path to of P
       Sort all paths by total underestimate, shortest first.
If G found -> success. Else -> failure.
]]
function _findPath( startSystem, endSystem, callback )
    print(2)
    print("looking for system at "..endSystem.gridPosition.i..", "..endSystem.gridPosition.j)
    local open = {
        {
            [1] = startSystem;
            ["underestimate"] = 0;
        }
    }
    local closed = {}
    local endi = endSystem.gridPosition.i
    local endj = endSystem.gridPosition.j
    local cycles = 0
    local totalCycles = 0
    
    repeat
        -- extract the first path from the open set.
        local path = open[1]
        table.remove( open, 1 )
        -- get the last system in the path.
        local system = path[#path]
        local i, j = system.gridPosition.i, system.gridPosition.j
        local s = "==== "
        for i, v in ipairs(path) do
            s = s .. ("(%s,%s) -> "):format( v.gridPosition.i, v.gridPosition.j )
        end
        print(s)
        -- check each connected system
        for i, connectedSystem in ipairs( system:getConnectedSystems() ) do
            -- check if we're looping onto an system on the path
            local looping = false
            for i, _system in ipairs(path) do
                if _system == connectedSystem then
                    looping = true
                    break
                end
            end
            if not looping then
                -- calculate an underestimate (manhatten distance) of the distance remaining to goal
                -- and add this to the distance travelled.
                local i1, j1 = connectedSystem.gridPosition.i, connectedSystem.gridPosition.j
                local travelled = math.abs( (i - i1) + (j - j1) )
                local remaining = math.abs( (i1 - endi) + (j1 - endj) )
                local underestimate = path.underestimate + travelled + remaining
                print( ("check system at %s, %s  (%s)"):format(i1,j1,underestimate) )
                -- copy path into a new path
                local newPath = {
                    ["underestimate"] = underestimate;
                }
                for i, v in ipairs(path) do
                    newPath[i] = v
                end
                -- append the connected system to the new path
                newPath[#newPath+1] = connectedSystem
                -- check all other paths that lead to the same system, and discard the new path if it has a larger estimate.
                local shouldDiscard = false
                for i, path in ipairs(open) do
                    if path[#path] == connectedSystem and path.underestimate < underestimate then
                        shouldDiscard = true
                        break
                    end
                end
                if not shouldDiscard then
                    -- insert the new path into the open set in ascending order of the underestimate.
                    for i = 1, #open+1 do
                        if i == #open+1 then
                            open[#open+1] = newPath
                        elseif open[i].underestimate >= newPath.underestimate then
                            table.insert( open, i, newPath )
                            break
                        end
                    end
                end
            end
        end
        -- We yield if too many cycles have past.
        cycles = cycles + 1
        totalCycles = totalCycles + 1
        if cycles >= MAX_CYCLES_PER_RESUME then
            cycles = 0
            coroutine.yield()
        end
    until #open == 0 or open[1][#open[1]] == endSystem or totalCycles >= MAX_CYCLES
    
    if #open == 0 then 
        callback( false )
        return false
    else
        print("FOUND IT!")
        s = "==== "
        for i, v in ipairs(open[1]) do
            s = s .. ("(%s,%s) -> "):format( v.gridPosition.i, v.gridPosition.j )
        end
        print(s)
        callback( open[1] )
        return open[1]
    end
end