PortController = class('PortController', EntityController)
PortController.instances = {}
-- requires
require "PortInputController"
require "PortRenderer"
-- States
PORT_UNCONNECTED    = 0
PORT_CONNECTED      = 1
PORT_OPEN           = 2
PORT_FORWARDED      = 3
--
-- upgrades
PORT_UPGRADE_NONE       = 0
PORT_UPGRADE_FIREWALL   = 1
--

function PortController:initialize( systemController, side, portRenderer )
    EntityController.initialize( self )
	self.systemController = systemController
	self.systemControllerSide = side
	self.packets = {}
	self.transferSpeed = 100
	self.forwardSpeed = 5
	self.upgrade = PORT_UPGRADE_NONE
	self.upgradeProperties = nil
	self.upgradable = true
	self.state = PORT_UNCONNECTED
	self.sisterPort = nil
	self.forwardPort = nil
	self.isMaster = false
	self.backdoor = false
	self.portRenderer = portRenderer
	self.portRenderer:setPortController( self )
	self.inputController = PortInputController( self )
	self.boundingBox = BoundingBox(0,0,10,10)
	self:calculatePosition()
	table.insert( PortController.instances, self )
end

function PortController:calculatePosition()
    local edge = self.systemController.hexagonShape:getEdgeBySide( self.systemControllerSide, 25 )
	local t = (1/6) + (1/5) * self.systemController:numberOfPortsOnSide( self.systemControllerSide )
    if self.isMaster then
        self.position.x, self.position.y = util.lerpXY( edge[1], edge[2], edge[3], edge[4], t )
    else
        local a, b, c, d = util.reverseXY(edge[1], edge[2], edge[3], edge[4])
        self.position.x, self.position.y = util.lerpXY( a, b, c, d, t )
    end
    self.boundingBox:placeAroundPosition( self.position )
end

function PortController:update( dt )
    --if true then return false end
    -- update input contoller.
    self.inputController:update( dt )
    -- update transfer progress of all packets.
    for i, packet in ipairs(self.packets) do
        packet.transferProgress = packet.transferProgress + (self.transferSpeed * dt)
        if packet.transferProgress >= 100 then
            self:deliverPacket( packet )
        end
    end
end

function PortController:connectToPort( port )
	self.sisterPort = port
	self.state = PORT_CONNECTED
	self.isMaster = true
	port.sisterPort = self
	port.state = PORT_CONNECTED
	port.isMaster = false
	self:calculatePosition()
	port:calculatePosition()
end

function PortController:forwardToPort( port )
    -- check port is in same system and not self.
    if port.systemController == self.systemController and port ~= self then
        self.forwardPort = port
        self.state = PORT_FORWARDED
        port.forwardPort = self
        port.state = PORT_FORWARDED
    end
end

function PortController:canTransferProgram( program )
    if self.state == PORT_CONNECTED then
        -- disallow enemy programs if firewall enabled.
        local programTeam = program:getTeam()
        if self.upgrade == PORT_UPGRADE_FIREWALL then
            if programTeam == self.upgradeProperties.team then
                return true
            else
                return false
            end
        end
        -- default when connected.
        return true
    end
    
    return false
end

function PortController:transferProgram( program )
    if not self.sisterPort then 
        error('PortController needs a sister port to transfer program.')
        return false
    end
    -- create new packet if none exists.
    local packet
    if #self.packets == 0 then
        packet = {
            transferProgress = 0,
            programs = {}
        }
        self.packets[1] = packet
    else
        packet = self.packets[ #self.packets ]
    end
    -- add program to packet.
    if packet then
        table.insert( packet.programs, program )
        -- tell sister port about transfer
        self.sisterPort:onPortTransferPacket( self, packet )
        return true
    else
        error( "packet == nil" )
        return false
    end
end

function PortController:deliverPacket( packet, key )
    -- check packet exists
    if not packet then
        error( "packet == nil" )
        return false
    end
    
    -- remove packet from packets array.
    self:removePacket( packet )
    -- tell sister port about the packet.
    self.sisterPort:onPortDeliveredPacket( self, packet )
end

function PortController:removePacket( packet )
    for i, testPacket in ipairs(self.packets) do
        if testPacket == packet then
            table.remove( self.packets, i )
            return true
        end
    end
    return false
end

function PortController:onPortTransferPacket( port, packet )
end

function PortController:onPortDeliveredPacket( port, packet )
    for i, program in ipairs(packet.programs) do
        self.systemController:addProgram( program )
        program:onExitPort( self, self.systemController, self.position )
    end
end

function PortController:setTransferSpeed( speed )
    self.transferSpeed = speed
    if self.state == PORT_CONNECTED then
        self.sisterPort.transferSpeed = speed
    end
end

function PortController:getSisterPort()
    return self.sisterPort
end

function PortController:getConnectedSystem()
    return self.sisterPort.systemController
end

function PortController:isConnected()
    return (self.state == PORT_CONNECTED) and self.sisterPort
end

function PortController:getPacketWithProgram( programController )
    for i, packet in ipairs(self.packets) do
        for _, program in ipairs(packet.programs) do
            if program == programController then
                return packet
            end
        end
    end
    return nil
end

function PortController:hasProgram( programController )
    for i, packet in ipairs(self.packets) do
        for _, program in ipairs(packet.programs) do
            if program == programController then
                return true
            end
        end
    end
    return false
end

function PortController:hasPacket( packet )
    for i, testPacket in ipairs(self.packets) do
        if testPacket == packet then
            return true
        end
    end
    return false
end

function PortController:setBackdoor( backdoor )
    self.backdoor = backdoor
    if backdoor then
        self.transferSpeed = self.transferSpeed / 4
    end
end

function PortController:isBackdoor()
    return self.backdoor
end