require 'EntityController'
require 'BoundingBox'
require 'ProgramRenderer'

ProgramController = class('ProgramController', EntityController)

STATE_APPROACH_PORT     = 1
STATE_APPROACH_WAYPOINT = 2
STATE_APPROACH_TARGET   = 3
STATE_STATIONARY        = 4
STATE_INSIDE_PORT       = 5
STATE_NAVIGATING_PATH   = 6

function ProgramController:initialize( inputController, systemController, programRenderer )
    EntityController.initialize( self )
    self.systemController = systemController
    self.position = vector( 0, 0 )
    self.boundingBox = BoundingBox(-10, -10, 10, 10)
    self.inputController = inputController
    self.inputController:setProgramController(self)
    if programRenderer then
        self.programRenderer = programRenderer
        self.programRenderer:setProgramController(self)
    end
    self.orderQueue = {}
    self.selected = false
    self.team = 0
    self.speed = 400
    self.state = STATE_STATIONARY
    self.listener = NotificationController.listen( "deselect_all_programs", nil, self.onDeselectAllPrograms, self )
end

function ProgramController:destroy()
    EntityController.destroy( self )
    NotificationController.killListener( self.listener )
end

function ProgramController:update( dt )
    self.inputController:update( dt )
    
    local order = self:popFirstInOrderQueue()
    if order then
        if order.orderType == 'move' then
            self.destination = vector( order.x, order.y )
        elseif order.orderType == 'port' then
            self.targetPort = order.portController
            self.destination = order.portController.position
            self.state = STATE_APPROACH_PORT
        end
    end
    
    if self.state == STATE_APPROACH_PORT then
        if vector.distance( self.position, self.targetPort.position ) < self.speed * dt then
            self:enterPort( self.targetPort )
            self.state = STATE_INSIDE_PORT
            self.destination = nil
        end
    end
    
    if self.pathToExternalSystem and #self.pathToExternalSystem > 0 then
        if not self.destinationSystem or self.systemController == self.destinationSystem then
            self.destinationSystem = self.pathToExternalSystem[1]
            table.remove( self.pathToExternalSystem, 1 )
            local ports = self.systemController:getPortsConnectedToSystem( self.destinationSystem )
            if #ports > 0 then
                local port = ports[ math.random(#ports) ]
                if port then
                    self:addPortOrder( port )
                end
            end
        end
    end
    
    if self.destination then
        local direction = vector.normalize(self.destination - self.position)
        self:moveBy( direction * dt * self.speed )
        if vector.distance( self.position, self.destination ) <= self.speed * dt then
            self.destination = nil
        end
    end
end

function ProgramController:getSystem()
    return self.systemController
end

function ProgramController:setSystem( systemController )
    self.systemController = systemController
end

function ProgramController:navigateToSystem( systemController )
    GridController.instance.pathfinder:findPath( self.systemController, systemController, function(path)
        if path then
            self.pathToExternalSystem = path
            self.destinationSystem = nil
        end
    end)
end

function ProgramController.setPathToExternalSystem( self, path )
    if path then
        self.pathToExternalSystem = path
        self.destinationSystem = nil
    end
end

function ProgramController:enterPort( portController )
    if portController:canTransferProgram( self ) then
        portController:transferProgram( self )
        self.systemController:removeProgram( self )
        self.systemController = nil
        self.portController = portController
        self.state = STATE_INSIDE_PORT
        self.inputController:halt()
    end
end

function ProgramController:onExitPort( portController, systemController, portPosition )
    local randomVec = vector( math.random(5, 20), math.random(5, 20) )
    self:setPosition( portPosition + randomVec )
    self.portController = nil
    self.systemController = systemController
    self.state = STATE_STATIONARY
    self.inputController:resume()
end

function ProgramController:position()
    return self.position
end

function ProgramController:setPosition( x, y )
    if type(x) == "table" then
        self.position = x:clone()
    else
        self.position = vector( x, y )
    end
    self.boundingBox:placeAroundPosition( self.position )
end

function ProgramController:moveBy( v )
    self.position = self.position + v
    self.boundingBox:placeAroundPosition( self.position )
end

function ProgramController:getBoundingBox()
    return self.boundingBox
end

function ProgramController:getSelected()
    return self.selected
end

function ProgramController:setSelected( selected, shouldNotify )
    if selected == true and shouldNotify then
        NotificationController.send("deselect_all_programs")
    end
    self.selected = selected
end

function ProgramController:onDeselectAllPrograms()
    self.selected = false
end

function ProgramController:addMoveOrder( x, y )
    local order = {
        orderType = 'move',
        x = x, y = y
    }
    self:addToOrderQueue( order )
end

function ProgramController:addAttackOrder( target )
    local order = {
        orderType = 'attack',
        target = target
    }
    self:addToOrderQueue( order )
end

function ProgramController:addAssistOrder( target )
    local order = {
        orderType = 'assist',
        target = target
    }
    self:addToOrderQueue( order )
end

function ProgramController:addPortOrder( portController )
    local order = {
        orderType = 'port',
        portController = portController
    }
    self:addToOrderQueue( order )
end

function ProgramController:addToOrderQueue( order )
    table.insert( self.orderQueue, order )
end

function ProgramController:popFirstInOrderQueue()
    local order = self.orderQueue[1]
    table.remove( self.orderQueue, 1 )
    return order
end

function ProgramController:setTeam( team )
    self.team = team
end

function ProgramController:getTeam()
    return self.team
end

function ProgramController:isEnemyWith( programController )
    return self.team ~= programController:getTeam()
end

function ProgramController:isAlliedWith( programController )
    return self.team == programController:getTeam()
end