-- requires
require 'EntityController'
require 'SystemRender'
require 'Hexagon'

SystemController = class('SystemController', EntityController)

function SystemController:initialize( gridController, i, j, hexagonShape, systemSpacing )
    EntityController.initialize( self )
    self.gridController = gridController
    self.gridPosition = { i = i, j = j }
	self.programs = {}
	self.ports = {}
	self.CPU = 0
	self.energy = 0
	self.memory = 0
	self.usedSides = {}
	self.hexagonShape = hexagonShape
	self.systemSpacing = systemSpacing
	self.connectedSystems = {}
	
	-- set even flag.
	local yOffset
    if self.gridPosition.i % 2 == 0 then
        self.isEven = true
        yOffset = hexagonShape.verticalOffset * systemSpacing
    else
        self.isEven = false
        yOffset = 0
    end
    
    -- set position.
    self.position.x = i * hexagonShape.horizontalOffset * systemSpacing
    self.position.y = (j * hexagonShape.verticalOffset   * systemSpacing * 2) + yOffset
    self.boundingBox = BoundingBox( 
        self.position.x + hexagonShape.aa.x, 
        self.position.y + hexagonShape.aa.y, 
        self.position.x + hexagonShape.bb.x,
        self.position.y + hexagonShape.bb.y
    )
    
	self.systemRender = SystemRender( self, hexagonShape, systemSpacing )
end

function SystemController:update( dt )
end

function SystemController:addProgram( program )
    table.insert( self.programs, program )
end

function SystemController:removeProgram( program )
    for k, program1 in ipairs( self.programs ) do
        if program1 == program then
            table.remove( self.programs, k )
            return true
        end
    end
    -- program not found.
    return false
end

function SystemController:createPort( system, side )
    local reverseSide = util.reverseHexDirection( side )
    local port = PortController:new( self, side, PortRenderer() )
    local sisterPort = PortController:new( system, reverseSide, PortRenderer() )
    if self:numberOfPortsOnSide( side ) == 0 then
        port:setBackdoor( true )
        sisterPort:setBackdoor( true )
    end
    port:connectToPort( sisterPort )
    table.insert( self.ports, port )
    table.insert( system.ports, sisterPort )
    
    if not self:isConnectedToSystem( system ) then
        table.insert( self.connectedSystems, system )
    end
    if not system:isConnectedToSystem( self ) then
        table.insert( system.connectedSystems, self )
    end
    
    self.usedSides[ side ] = true
    system.usedSides[ reverseSide ] = true
end

function SystemController:createPortOnSide( side )
    local system = self:findNeighborSystem( side )
    if system then
        self:createPort( system, side )
        return system
    else
        return false
    end
end

function SystemController:numberOfPortsOnSide( side )
    local count = 0
    for i, port in ipairs(self.ports) do
        if port.systemControllerSide == side then
            count = count + 1
        end
    end
    
    return count
end

function SystemController:getPortConnectedToSystem( system )
    for i, port in ipairs(self.ports) do
        if port:isConnected() and port:getConnectedSystem() == system then
            return port
        end
    end
end

function SystemController:getPortsConnectedToSystem( system )
    local ports = {}
    for i, _port in ipairs(self.ports) do
        if _port:isConnected() and _port:getConnectedSystem() == system then
            ports[#ports+1] = _port
        end
    end
    return ports
end

function SystemController:getPortsOnSide( side )
    local ports = {}
    for i, _port in ipairs(self.ports) do
        if _port.systemControllerSide == side then
            ports[#ports+1] = _port
        end
    end
    
    return ports
end

function SystemController:selectRandomPortOnSide( side )
    local ports = self:getPortsOnSide( side )
    return ports[ math.random(#ports) ]
end

function SystemController:getConnectedSystems()
    return self.connectedSystems
end

function SystemController:isConnectedToSystem( system )
    for i, connectedSystem in ipairs(self.connectedSystems) do
        if connectedSystem == system then
            return true
        end
    end
    return false
end

function SystemController:listOfNeighbors()
    local neighbors = {}
    for i = 0, 5 do
        local neighbor = self:findNeighborSystem( i )
        if neighbor then
            table.insert( neighbors, { system = neighbor, side = i } )
        end
    end
    
    return neighbors
end

function SystemController:findNeighborSystem( side )
    local i = self.gridPosition.i
    local j = self.gridPosition.j
    
    return self.gridController:findNeighborSystem( i, j, side )
end

function SystemController:isPointInside( px, py )
    local q = vector(px, py) - self.position
    return self.hexagonShape:isPointInside( q.x, q.y )
end