-- Program spec.
require 'specs.init'
require 'GridController'
require 'ProgramController'
require 'ProgramInputController'
require 'PortController'

createMatcher("is_class", "'%s' to be class '%s'", 2, function(a,b) return instanceOf(b,a) end )

describe["A Program"] = function()
    before = function()
        MockGridController = Mock:new()
        MockGridController.isSystemUnderMouse:returns( true )
        GridController.instance = MockGridController
        MockSystemController = Mock:new()
        MockSystemController.position = vector(0,0)
        program = ProgramController( ProgramInputController(), MockSystemController )
    end
    
    after = function()
        GridController = nil
        program:destroy()
    end
    
    it["should belong to a system when initialized"] = function()
        expect( program:getSystem() ).should_be( MockSystemController )
    end
    
    it["should be able to change system"] = function()
        local MockSystemControllerB = MockSystemController:clone()
        program:setSystem( MockSystemControllerB )
        expect( program:getSystem() ).should_be( MockSystemControllerB )
    end

    it["should have x and y co-ordinates"] = function()
        expect( program.position ).is_type( "table" )
        expect( program.position.x ).is_type( "number" )
        expect( program.position.y ).is_type( "number" )
    end

    it["should be able to change co-ordinates"] = function()
        program:setPosition( 100, 200 )
        expect( program.position.x ).should_be( 100 )
        expect( program.position.y ).should_be( 200 )
    end
    
    it["has a bounding box"] = function()
        local boundingBox = program:getBoundingBox()
        expect( boundingBox ).is_not_nil()
        expect( boundingBox ).is_class( BoundingBox )
    end
    
    it["should update the bounding box when it changes position"] = function()
        local boundingBox = program:getBoundingBox()
        boundingBox:set(0, 0, 20, 20)
        program:setPosition( 100, 100 )
        expect( boundingBox.x1 ).should_be( 90 )
        expect( boundingBox.y1 ).should_be( 90 )
        expect( boundingBox.x2 ).should_be( 110 )
        expect( boundingBox.y2 ).should_be( 110 )
    end
    
    describe["when entering a port"] = function()
        before = function()
            MockPortController = Mock:new()
            MockPortController.canTransferProgram:returns( true )
            program:enterPort( MockPortController )
        end
        
        it["should be removed from the system"] = function()
            expect( program.systemController ).is_nil()
        end
        
        it["should belong to the port"] = function()
            expect( program.portController ).should_be( MockPortController )
        end
    end
    
    describe["Input interaction"] = function()
        before = function()
           input = program.inputController
           program:getBoundingBox():set( 0, 0, 20, 20 )
           worldMousePos.x = 10
           worldMousePos.y = 10
           love.mouse.resetDown()
        end
        
        it["should be in 'mouse over' state when mouse is over and no mouse buttons are pressed"] = function()
            program:update(1)
            expect( input:getMouseOverState() ).should_be( MOUSE_OVER )
        end
        
        it["should be in 'mouse off' state when mouse if off"] = function()
            worldMousePos.x = 30
            program:update(1)
            expect( input:getMouseOverState() ).should_be( MOUSE_OFF )
        end
        
        it["should be in 'mouse down' state when mouse is over and the left, middle or right button is being pressed"] = function()
            love.mouse.down.l = true
            program:update(1)
            expect( input:getMouseOverState() ).should_be( MOUSE_DOWN )
            
            love.mouse.resetDown()
            love.mouse.down.r = true
            program:update(1)
            expect( input:getMouseOverState() ).should_be( MOUSE_DOWN )
            
            love.mouse.resetDown()
            love.mouse.down.m = true
            program:update(1)
            expect( input:getMouseOverState() ).should_be( MOUSE_DOWN )
        end
        
        it["should be selected when clicked with left click"] = function()
            love.mouse.down.l = true
            input:onEnterMouseDown( program )
            expect( program:getSelected() ).should_be( true )
        end
        
        it["should be deselected when another program is selected"] = function()
            local programB = ProgramController( ProgramInputController(), MockSystemController )
            program:setSelected( true, true )
            programB:setSelected( true, true )
            expect( program:getSelected() ).should_be( false )
            expect( programB:getSelected() ).should_be( true )
            programB:destroy()
        end
        
        it["should be deselected when empty space is left clicked"] = function()
            program:setSelected( true )
            worldMousePos.x = 30
            worldMousePos.y = 30
            input:onMouseButtonPressed( 0, 0, 'l' )
            expect( program:getSelected() ).should_be( false )
        end
        
        describe["Orders"] = function()
            before = function()
                program:setSelected( true )
                program_b = ProgramController( ProgramInputController(), MockSystemController )
                worldMousePos.x = 100
                worldMousePos.y = 100
            end
            
            after = function()
                program_b:destroy()
            end
        
            it["should queue a move order when right clicked over empty space"] = function()
                input:onMouseButtonPressed( 0, 0, 'r' )
                local order = program:popFirstInOrderQueue()
                expect( order ).should_equal( {orderType = 'move', x = 100, y = 100} )
            end
        
            it["should queue an attack order when right clicked over enemy program"] = function()
                program_b:setTeam( 2 )
                program:setTeam( 1 )
                program_b.inputController.mouseOverState = MOUSE_OVER
                input:onMouseButtonPressed( 0, 0, 'r' )
                local order = program:popFirstInOrderQueue()
                expect( order ).should_equal( { orderType = 'attack', target = program_b} )
            end
        
            it["should queue an assist order when right clicked over allied program"] = function()
                program_b:setTeam( 1 )
                program:setTeam( 1 )
                program_b.inputController.mouseOverState = MOUSE_OVER
                input:onMouseButtonPressed( 0, 0, 'r' )
                local order = program:popFirstInOrderQueue()
                expect( order ).should_equal( { orderType = 'assist', target = program_b } )
            end
            
            it["should not queue an order if right clicked outside the system"] = function()
                MockGridController.isSystemUnderMouse:returns( false )
                worldMousePos.x = -100
                worldMousePos.y = -100
                input:onMouseButtonPressed( 0, 0, 'r' )
                local order = program:popFirstInOrderQueue()
                expect( order ).is_nil()
            end
            
            describe["when right clicking a port"] = function()
                before = function()
                    MockPortController = Mock:new()
                    MockPortController.systemController:returns( MockSystemController )
                    MockPortController.position = vector( 10, 10 )
                    program:addPortOrder( MockPortController )
                end
                
                it["should queue a port order"] = function()
                    local order = program:popFirstInOrderQueue()
                    expect( order ).should_equal( { orderType = "port", portController = MockPortController } )
                end
                
                it["should have the port's position as it's destination"] = function()
                    program:update(1)
                    expect( program.destination ).should_equal( MockPortController.position )
                end
            end
        end
    end
end